# -*- coding: utf-8 -*-
"""
 (C) Jason Kinser 2022.
 This code is intended for non-commercial use in courses at George Mason Univer
 Commercial and/or non-academic use prohibited without written permission from the author
 Modified by MVC 2/2/25
"""
import numpy as np
import matplotlib.pyplot as plt

# ############################################
# Matrix version
# ############################################

def InitVirusMx(NPeople = 100, space=99):
    M = np.zeros((NPeople,6)) # x1,x2, v1,v2, a1,a2
    # расставим людей в мире
    M[:,:2] = np.random.ranf((NPeople,2))*space # random_sample
    M[:,2:4] = np.random.ranf((NPeople,2))*5+5
    #
    I = np.zeros((NPeople,2),int) # запас здоровья
    I[:,0] = np.random.randint(50,150,NPeople)
    I[:,1] = np.random.randint(-5,100,NPeople)
    #
    B = np.zeros((NPeople,3), bool) # alive, infected, immune
    B[:,0] = True # все живы
    B[:,1] = False # не заражен
    B[:,2] = False # без иммунитета
    return M, B, I

def CollideM(M,B,me, him):
    #print(me,him)
    alpha = np.arctan2(M[him,1] - M[me,1], M[him,0] - M[me,0])
    R = np.array(((np.cos(-alpha),-np.sin(-alpha)),(np.sin(-alpha),np.cos(-alpha))))
    vrot = R.dot(M[me,2:4])
    wrot = vrot * np.array((1,-1))
    R = np.array(((np.cos(alpha),-np.sin(alpha)),(np.sin(alpha),np.cos(alpha))))
    w = R.dot(wrot)
    M[me,2:4] = w + 0
    # перенос болезни
    if B[him,1] and not B[me,2]:
       B[me,1] = True # (if np.random.random()>0.25 else False)

def IterateM(M,B,I,space=300):
    dt = 0.1
    X = Y = space
    alive = B[:,0].nonzero()[0] # счет живых
    for me in alive:
        # перемещение в мире
        M[me,:2] = M[me,:2] + M[me,2:4]*dt + 0.5*M[me,4:6]*dt**2
        # проверка границ мира
        if M[me,0]<0: 
            M[me,0] *= -1;  M[me,2] *= -1
        if M[me,0]>X: 
            M[me,0] = X - (M[me,0] -X)
            M[me,2] *= -1
        if M[me,1]<0: 
            M[me,1] *= -1;  M[me,3] *= -1
        if M[me,1]>Y: 
            M[me,1] = Y - (M[me,1] -Y)
            M[me,3] *= -1
        # проверка уровня здоровья
        if B[me,1] and I[me,0] > I[me,1]:
           I[me,0] -= 1
        # излечение
        if I[me,0]<=I[me,1]:
           B[me,1:3] = False, True
        # вычисление расстояния до других
        them = list(alive)
        them.remove(me)
        them = np.array(them)
        dist = np.sqrt(((M[me,:2]-M[:,:2])**2).sum(1))
        # дистанция заражения <5
        closendx = (dist<4).nonzero()[0]
        # проверка столкновений людей
        for i in closendx:
            if i in them:
                CollideM(M,B,me,i)
                CollideM(M,B,i,me)
        # проверка смерти
        ndx = (I[:,0]<=0).nonzero()[0]
        B[ndx] = False,True,False
####
def RunMatrixVirus(NPeople=500,N_ITERS=250):
    space = 250
    M, B, I = InitVirusMx(NPeople,(space-1))
    B[:10,1] = True # 10 персон больны
    inf = np.zeros(N_ITERS)
    aliv = np.zeros(N_ITERS)
    immun = np.zeros(N_ITERS)
    clean = np.zeros(N_ITERS)
    for step in range(N_ITERS):
        IterateM(M, B, I, space)
        inf[step] = int(B[:,1].sum())
        aliv[step] = int(B[:,0].sum())
        immun[step] = int(B[:,2].sum())
        clean[step] = aliv[step]-inf[step]-immun[step]
        if step%10==0: print(step, end='.')
        elif step%49==0: print()
    #
    return inf, clean, immun

def main(variant=0):
    inf, aliv, health = RunMatrixVirus(1000,500)
    plt.plot(inf,color='red')
    plt.plot(aliv,color='green')
    plt.plot(health,color='yellow')
    plt.show()
    #

main()
