# -*- coding: utf-8 -*-
"""
 Created (C) Jason Kinser 2022
 This code is intended for non-commercial use in courses at George Mason Univ.
 Commercial and/or non-academic use prohibited without written permission from the author.
 Modified by MVC 2/4/25
"""
import numpy as np
import matplotlib.pyplot as plt

SEED = 916421
SPACES = 200
POPULATION = 500
VELOCOEF = 20
ASCOEF = 0.1
INI_ILL = 10
XLim = SPACES; YLim=XLim+50
HEALTH_MID = 75
R_COMM = 4
IMUNOPOWER = 0.5
N_VAKCIN = 5
CollideCount =0
people, Inf_list =[],[]
Inf_count =0

### это класс для персоны
class Person:
    global VELOCOEF, ASCOEF, POPULATION, SPACES, HEALTH_MID, Inf_count
    
    def __init__(self):
        self.x = np.random.rand(2)*(SPACES-1)  # координаты x-y
        self.v = np.random.rand(2)*VELOCOEF +1  # скорость x-y
        self.a = np.zeros(2)+ASCOEF  # ускорение x-y
        self.alive = True
        self.immune = False
        self.infected = False
        self.health = np.random.randint(HEALTH_MID-30,HEALTH_MID+30) # текущий уровень здоровья
        self.bottom = np.random.randint(-1,self.health) # критический уровень здоровья

    def __str__( self ):
        st = '';  temps = vars(self)
        for item in temps:  st += f'{item}: {temps[item]}\n'
        return st

    def Move(self, V, H, dt = 0.1):
        self.x[0] = self.x[0] + self.v[0]*dt*np.random.choice([-1,1]) + 0.5*self.a[0]*dt*dt
        self.x[1] = self.x[1] + self.v[1]*dt + self.a[1]*dt*dt
        if self.x[0] < 0:
           self.x[0] *= -1;  self.v[0] *= -1
        if self.x[0] > H:
           self.x[0] = H - 1 #(self.x[0]-H)
           self.v[0] *= -1
        if self.x[1] < 0:
           self.x[1] *= -1;  self.v[1] *= -1
        if self.x[1] > V:
           self.x[1] = V - 1 #(self.x[1]-V)
           self.v[1] *= -1

    def LowerHealth(self):
        # снижение здоровья от болезни
        if self.infected and self.health >self.bottom:
            self.health -= 1
        # если здоровье закончилось то вычеркиваем
        self.alive = (self.health > 0)

    def Cured(self):
        global Inf_count
        # можем вылечиться
        if self.health <= self.bottom:
           self.infected = False
           Inf_count -=1
           self.immune = True
           self.bottom -= 5
        # можем поднять здоровье если есть иммунитет
        if self.immune:
           #self.health = (self.health+1) if np.random.choice([True, False]) else self.health
           self.health = self.health + np.random.choice([1,0])

    def Distance(self, you):
        return np.sqrt(((self.x-you.x)**2).sum())

    def Collide(self,you):
        # проверка столкновения персон
        ypart = you.x[1] - self.x[1]
        xpart = you.x[0] - self.x[0]
        alpha = np.arctan2(ypart, xpart)
        R = np.array(((np.cos(-alpha),-np.sin(-alpha)),
                      (np.sin(-alpha),np.cos(-alpha))))
        vrot = R.dot(self.v)
        wrot = vrot * np.array((1,-1))
        R = np.array(((np.cos(alpha),-np.sin(alpha)),
                      (np.sin(alpha),np.cos(alpha))))
        w = R.dot(wrot)
        self.v = w + 0.05
        # передача болезни
        if you.infected:
           if self.immune:
              self.infected = (np.random.ranf() > IMUNOPOWER)
              self.immune = not self.infected
           else:
              self.infected = True


### создание мира
def CreateWorld(NPeople=100):
    people = [Person() for _ in range(NPeople)]
    # задаем несколько больных
    for p in people[:INI_ILL]:  p.infected=True
    return people

# выполнение прохода по модели
def Iterate():
    global R_COMM, CollideCount, XLim, YLim, people, Inf_count
    #
    for me in people:
        if me.alive:
           me.Move(YLim,XLim,dt=0.11)
           me.LowerHealth()
           me.Cured()
           # проверка встречи!
           for you in people:
              if you.alive and me != you:
                  if me.Distance(you) < R_COMM:
                     CollideCount +=1
                     me.Collide(you)
                     you.Collide(me)
        if me.infected: Inf_count +=1

# запуск модели
def main(variant=0):
    global XLim, YLim, POPULATION, people, Inf_list, Inf_count
    
    # собираем данные
    def CollectData(ix):
      nonlocal LivInfIm, N_ITERS, pos_liv
      #
      pos_liv[2,ix], pos_liv[1,ix], pos_liv[0,ix] = [],[],[]
      for p in people:
          if p.alive:
             LivInfIm[0,ix] +=1
             if p.immune:
                LivInfIm[2,ix] +=1
                pos_liv[0,ix].append(p.x.copy())
             elif p.infected:
                LivInfIm[1,ix] +=1
                pos_liv[1,ix].append(p.x.copy())
             else:      #if not p.immune and not p.infected:
                pos_liv[2,ix].append(p.x.copy())

    # рисуем графики
    def plotgrafiks():
        nonlocal LivInfIm, N_ITERS, pos_liv
        fig, axs = plt.subplots(1,2, figsize=(12,7))
        fig.tight_layout()
        # укажем размеры по осям
        axs[0].set(xlim=(0,XLim), ylim=(0,YLim))
        axs[1].set_title(f"число контактов {CollideCount}")
        axs[1].step(range(N_ITERS),LivInfIm[1], color='red')
        axs[1].step(range(N_ITERS),LivInfIm[2], color='black')
        axs[1].step(range(N_ITERS),(LivInfIm[0]-LivInfIm[1]-LivInfIm[2]), color='green')
        # запустим отображение динамики популяции
        for i in range(N_ITERS):
            axs[0].clear()
            axs[0].set(xlim=(0,XLim), ylim=(0,YLim), autoscale_on=False, title= f"{i} шагов" )
            # выводим данные - неболевшие
            if len(list(pos_liv[2,i])) >0:
               xx,yy = list(zip(*pos_liv[2,i]))
               axs[0].scatter(xx,yy, color='lime')
            # выводим данные - больные
            if len(list(pos_liv[1,i])) >0:
               xx1,yy1 = list(zip(*pos_liv[1,i]))
               axs[0].scatter(xx1,yy1, color='red', marker='^')
            # выводим данные - имунные
            if len(list(pos_liv[0,i])) >0:
              xx2,yy2 = list(zip(*pos_liv[0,i]))
              axs[0].scatter(xx2,yy2, color='black', marker=(5,2))
            # подкраска "жалюзи"
            axs[1].plot([i,i],[0,POPULATION-1], alpha=0.25, c='grey', lw=1)
            # отобразим и повторим
            plt.pause(0.1) # надо паузить для анимации
        plt.show()
    
    np.random.seed(SEED)
    # длительность наблюдения мира
    N_ITERS=250
    # создаем мир персон
    people = CreateWorld(POPULATION)
    # массив счетчиков = здоровые-больные-имунные
    LivInfIm = np.zeros((3,N_ITERS), dtype=int)
    # массив популяций для графика
    pos_liv=np.ones((3,N_ITERS), dtype=list)
    # запускаем модель,сбор данных,график
    for step in range(N_ITERS):
        Iterate()
        CollectData(step)
        if step%10==0: print(step, end='..')
        elif step%49==0: print()
    plotgrafiks()

main()
